import pymongo
import pymongo.errors
import trafilatura
import trafilatura.feeds
import trafilatura.sitemaps
import trafilatura.spider
import trafilatura.utils
import trafilatura.external
import sys
import courlan
import urllib
from datetime import datetime
import click
import logging as LOGGER
import os
import pprint
import re
import time
import collections
import math
import random
import hashlib

LANGUAGE= os.getenv("SUCKER_LANGUAGE","sk")
DOMAIN = os.getenv("SUCKER_DOMAIN","sk")
BATCHSIZE=int(os.getenv("SUCKER_BATCHSIZE","10"))
CONNECTION=os.getenv("SUCKER_CONNECTION","mongodb://root:example@localhost:27017/")
DBNAME=os.getenv("SUCKER_DBNAME","crawler")
MINFILESIZE=300
MAXFILESIZE=10000000
MINTEXTSIZE=200
CHECK_PARAGRAPH_SIZE=150
TEXT_TRASH_SIZE=200
TEXT_TRASH_RATIO=0.6
DISCOVER_LINK_RATIO = 0.3
SAMPLE_SET_SIZE =10000
CLASSIFIER_SET_SIZE = 200
STOP_PATHS=["xml","rss","login","admin"]

def split_train(res):
    trainset = []
    testset = []
    for i,item in enumerate(res):
        if i % 10 == 0:
            testset.append(item)
        else:
            trainset.append(item)
    return trainset,testset

def calculate_checksums(text):
    """
    @return fingerprints of a paragraphs in text. Paragraphs are separated by a blank line
    """
    checksums = []
    sizes = []
    hval = 0
    hsz = 0
    sz = 0
    for c in text:
        cv = ord(c)
        sz += 1
        if cv > 64:  # ignore non-ascii
            hval += (hval << 3) + cv
            zv = hval >> 31
            hval &= 0x7fffffff
            hval += zv
            hsz += 1
        if c == "\n" and hsz > 0:
            if hsz > CHECK_PARAGRAPH_SIZE:
                checksums.append(hval)
                sizes.append(sz)
            sz = 0
            hsz = 0
    if hsz > CHECK_PARAGRAPH_SIZE:
        checksums.append(hval)
        sizes.append(sz)
    return checksums, sizes

def is_robot_good(link,rules):
    # check robots.txt rules
    if rules is not None and not rules.can_fetch("*", link):
        return False
    return True

def is_link_good(link):
    r = courlan.check_url(link,strict=True,language=LANGUAGE)
    if r is None:
        return None
    llink,lhostname = r
    paths = set(llink.split("/"))
    for item in STOP_PATHS:
        if item in paths:
            return None

    #print(llink,lhostname)
    # hostname rules
    if not lhostname.endswith(DOMAIN):
        LOGGER.debug("bad hostname")
        return None
    if courlan.is_not_crawlable(llink):
        LOGGER.debug("not crawlable")
        return None
    return llink

def get_link_doc(link:str,status="frontlink")->dict:
    r  = courlan.check_url(link)
    assert r is not None
    link,host = r
    domain = courlan.extract_domain(link)
    return {"url":link,"host":host,"domain":domain,"status":status,"created_at":datetime.utcnow()}


def fetch_page(link:str)->(str,str):
    print("fetching:::::")
    print(link)
    final_link = link
    response = trafilatura.fetch_url(link,decode=False)
    time.sleep(2)
    html = None
    if response is not None :
        good = True
        if response.status != 200:
            good = False
            LOGGER.error('not a 200 response: %s for URL %s', response.status, url)
        elif response.data is None or len(response.data) < MINFILESIZE:
            LOGGER.error('too small/incorrect for URL %s', link)
            good = False
        # raise error instead?
        elif len(response.data) > MAXFILESIZE:
            good = False
            LOGGER.error('too large: length %s for URL %s', len(response.data), link)
        if good:
            html = trafilatura.utils.decode_response(response) 
        if html is not None:
            html, final_link = trafilatura.spider.refresh_detection(html, final_link)
            # is there a meta-refresh on the page?
            if final_link is None:  # malformed or malicious content
                html = None

        final_link = courlan.normalize_url(final_link)
    return final_link,html

def fetch_robot(base_url:str)->urllib.robotparser.RobotFileParser:
    try:
        rawrules = trafilatura.fetch_url("https://"+ base_url + "/robots.txt")
        #print(rawrules)
        rules = urllib.robotparser.RobotFileParser()
        rules.parse(rawrules.split("\n"))
        LOGGER.info('got robots')
    except Exception as exc:
        LOGGER.error('cannot read robots.txt: %s', exc)
        rules = None
    # exceptions happening here
    return rules


def extract_pages(link_batch:list,responses:list)->list:
    out = []
    for original_link,(final_link,html) in zip(link_batch,responses):
        doc = None
        assert original_link is not None
        if html is not None:
            doc = trafilatura.bare_extraction(html,url=final_link,with_metadata=True,include_formatting=False,target_language=LANGUAGE,favor_precision=True)
            if doc is not None:
                if not "text" in doc or len(doc["text"]) < MINTEXTSIZE:
                    # text too small
                    doc = None

        out.append((original_link,final_link,html,doc))
    return out

def set_content_checksums(doc):
    text = doc["text"]
    checksums,sizes = calculate_checksums(text)
    doc["text_size"] = len(text)
    doc["text_md5"] = hashlib.md5(text.encode("utf8")).hexdigest()
    doc["paragraph_checksums"] = checksums
    doc["paragraph_sizes"] = sizes
    doc["paragraph_sizes_sum"] = sum(sizes)
    end_sentence_marker = re.compile("\w[\.]")
    sentences = 0
    for item in re.finditer(end_sentence_marker,text):
        t = item.group(0)
        if t[0].islower():
            sentences += 1
    doc["sentences_count"] = sentences

def index_pages(db,hostname,extracted_pages):
    linkcol = db["links"]
    htmlcol = db["html"]
    contentcol = db["content"]
    checkcol = db["check"]
    links = []
    # stats of the batch
    good_document_count = 0
    document_count = 0
    text_size = 0
    good_text_size = 0
    original_text_size = 0
    for original_link,final_link,html,doc in extracted_pages:
        document_count += 1
        state = "good"
        link = original_link
        if original_link != final_link:
            print(original_link,final_link)
            linkcol.update_one({"url":original_link},{"$set":{"status":"redirect"}})
            link = final_link
        if html is None:
            state = "html_error"
        elif doc is None:
            state = "content_error"
        if doc is not None:
            set_content_checksums(doc)
            tsz = doc["text_size"]
            text_size += tsz
            psz = doc["paragraph_sizes_sum"]
            if tsz < TEXT_TRASH_SIZE or psz/tsz < TEXT_TRASH_RATIO:
                state = "small"
        # check copy
        if state == "good":
            origsz = 0
            for chs,paragraph_size in zip(doc["paragraph_checksums"],doc["paragraph_sizes"]):
                # index paragraph checksums
                nd = checkcol.find_one({"_id":chs})
                if nd is None:
                    origsz += paragraph_size

            if (1 - (origsz / tsz)) > TEXT_TRASH_RATIO:
                state = "copy"
            original_text_size += origsz
            print(origsz)
        if state == "good":
            good_document_count += 1
            good_text_size += doc["text_size"]
            htdoc = get_link_doc(link,state)
            htdoc["html"] = html
            htdoc["html_size"] = len(html)
            htdoc["html_md5"]= hashlib.md5(html.encode("utf8")).hexdigest()
            # can be revisited - upsert
            del htdoc["url"]
            htmlcol.update_one({"url":link},{"$set":htdoc},upsert=True)
            doc.update(get_link_doc(link,"good"))
            # todo extract links
            print(link,doc)
            del doc["url"]
            contentcol.update_one({"url":link},{"$set":doc},upsert=True)
            for chs in doc["paragraph_checksums"]:
                try:
                    checkcol.insert_one({"_id":chs})
                except pymongo.errors.DuplicateKeyError as err:
                    pass

        linkdoc = get_link_doc(link,state)
        del linkdoc["url"]
        linkcol.update_one({"url":link},{"$set":linkdoc})
    batchdoc = {
      "host": linkdoc["host"],
      "domain": linkdoc["domain"],
      "created_at":  datetime.utcnow(),
      "good_document_count":good_document_count,
      "document_count":document_count,
      "text_size":text_size,
      "good_text_size":good_text_size,
      "original_text_size":original_text_size,
      "batch_size": BATCHSIZE,
      "average_fetch_characters": text_size / BATCHSIZE,
    }
    db["batches"].insert_one(batchdoc)
    print(batchdoc)

from bs4 import BeautifulSoup
import urllib.parse
import os.path

def get_bs_links(link,html):
    # Extrakcia linkov zo stranky
    bs = BeautifulSoup(html, "lxml")
    base = link
    if bs.base is not None and "href" in bs.base.attrs:
        base = bs.base["href"]
    base = urllib.parse.urlparse(courlan.normalize_url(base))

    external_links = set()
    internal_links = set()
    # Normalizacia linkov
    for l in bs.find_all("a", href=True):
        if "rel" in l.attrs and l.attrs["rel"] == "nofollow" or "nofollow" in l.attrs:
            continue
        href = l["href"]
        try:
            parsed = urllib.parse.urlparse(courlan.normalize_url(href))
            netloc = parsed.netloc
            path = os.path.normpath(parsed.path)
            scheme = parsed.scheme
            # internal link
            if parsed.netloc == "":
                scheme = base.scheme
                netloc = base.netloc
                if not parsed.path.startswith("/"):
                    path = os.path.normpath(base.path +"/" + path)
            if not scheme.startswith("http"):
                continue
            if path.startswith("/"):
                path = path[1:]
            if path.endswith(")"):
                # javascript
                continue
            external = True
            if parsed.netloc == base.netloc:
                external = False
            href = urllib.parse.urlunparse((scheme,netloc,path,"","",""))
            href = courlan.normalize_url(href)
            if external:
                external_links.add(href)
            else:
                internal_links.add(href)
        except ValueError as err:
            print(err)
            pass
    return internal_links,external_links

def extract_links(link_batch:list,responses:list,hostname:str,rules,default_status="frontlink")->list:
    links = {}
    badrobot = 0
    for original_link,(final_link,html) in zip(link_batch,responses):
        status = default_status
        if html is None or len(html) < 256:
            continue
        internal_links, external_links = get_bs_links(final_link,html)
        #external_links = courlan.extract_links(html,final_link,external_bool=True,language=LANGUAGE)
        #internal_links = courlan.extract_links(html,final_link,external_bool=False,language=LANGUAGE)
        #print(extracted_links)
        for link in internal_links:
            if not is_robot_good(link,rules):
                badrobot += 1
                continue
            status = str(default_status)
            #print(link,status)
            links[link] = status
    outlinks = []
    badlink = 0
    for link,status in links.items():
        link = is_link_good(link)
        if link is None:
            badlink += 1
            continue
        outlinks.append((link,status))
    print(f"{len(links)} total links, {badrobot} badrobot {badlink} badlinks")
    return outlinks

def index_links(db,extracted_links):
    linkcol=db["links"]
    for link,status in extracted_links:
        if not is_link_good(link):
            continue
        if status == "frontlink" or status == "backlink":
            doc = get_link_doc(link,status)
            try:
                linkcol.insert_one(doc)
                # dont overwrite
            except pymongo.errors.DuplicateKeyError as ex:
                pass
        else:
            print("updating " + link,status)
            linkcol.update_one({"url":link},{"$set":{"status":status,"updated_at":datetime.utcnow()}})

def get_link_features(link):
    a, urlpath = courlan.get_host_and_path(link)
    features = re.split("[/?&]",urlpath)
    #features = re.split("[/?-_=]",urlpath)
    res = []
    for i,feature in enumerate(features):
        if len(feature) < 1:
            continue
        feature = re.sub("[0-9]","*",feature)
        res.append(str(i)+ "-" + feature)
    if len(res) < 2:
        return None
    res = res[:-1]
    return res

class LinkClassifier:
    def __init__(self):

        self.goodcounter = collections.Counter()
        self.badcounter = collections.Counter()
        self.good_count = 0
        self.bad_count = 0
        self.alpha = 0.001

    def train(self,links):
        for i,item in enumerate(links):
            link = item["url"]
            state = item["status"]
            cl = 0
            if state == "good":
                cl = 1
            print(cl,state,link)
            features = get_link_features(link)
            if features is None:
                continue
            lf = len(features)
            if state == "good":
                for feature in features:
                    self.good_count += 1
                    self.goodcounter[feature] += 1
            else:
                for feature in features:
                    self.bad_count += 1
                    self.badcounter[feature] += 1
        self.bdictsize = len(self.badcounter)
        self.gdictsize = len(self.goodcounter)

    def test(self,testset):
        # eval
        gg = 0
        true_positive = 0
        positive = 0
        false_negative = 0
        for item  in testset:
            l = item["url"]
            cl = 0
            if item["status"] == "good":
                cl = 1
            pcp = self.classify(l)
            r = 0
            if pcp > 0:
                r = 1
            if cl == 1:
               if r == 1:
                 true_positive += 1
               positive += 1
            if r == 1 and cl == 0:
                    false_negative += 1
            if r == cl:
                gg += 1
            else:
                print("MISS",l,cl,pcp)
        print(len(testset))
        print("Precision: {}, Recall: {}".format(true_positive/positive,true_positive/(true_positive+false_negative)))
        print("Accuracy:")
        acc = gg / len(testset)
        print(acc)
        return acc

    def classify(self,link):
        if self.good_count == 0 or self.bad_count == 0:
            return random.uniform(-0.1,0.1)
        features = get_link_features(link)
        res = 0
        gp = math.log(self.good_count) - math.log(self.good_count + self.bad_count)
        bp = math.log(self.bad_count) - math.log(self.good_count + self.bad_count)
        if features is None:
            return math.exp(gp) - math.exp(bp)
        gcc = math.log(self.gdictsize * self.alpha + self.good_count)
        bcc = math.log(self.bdictsize * self.alpha + self.bad_count)
        goodprob = 0
        badprob = 0
        for feature in features:
            g = math.log((self.goodcounter[feature] + self.alpha)) - gcc
            goodprob += g
            b = math.log(self.badcounter[feature] + self.alpha) - bcc
            badprob += b
            print(feature,g,b)
        pa = math.exp(goodprob + gp)
        pb = math.exp(badprob + bp)
        return pa - pb #+ random.uniform(-0.001,0.001)


def get_links(db,hostname,status,batch_size):
    linkcol = db["links"]
    res = linkcol.find({"host":hostname,"status":status},limit=batch_size)
    links = []
    for item in res:
        links.append(item["url"])
    print("Got {} {}".format(len(links),status))
    return links



def fetch_sitemap_links(start_link):
    out = []
    navigation_links = trafilatura.sitemaps.sitemap_search(start_link,target_lang=LANGUAGE)
    for link in navigation_links:
        out.append((link,"frontlink"))
    print("Fetched {} sitemap links".format(len(out)))
    return out

def fetch_front_links(start_link,rules):
    start_link,hostname = courlan.check_url(start_link)
    response = fetch_page(start_link)
    extracted_links = extract_links([start_link],[response],hostname,rules,"frontlink")
    print("Fetched {} frontlinks".format(len(extracted_links)))
    return extracted_links



def link_summary(db,hostname):
    linkcol = db["links"]
    #res = linkcol.distinct("hostname",{"hostname":hostname})
    res = linkcol.aggregate([
        {"$match":{"host":hostname}},
        {"$group":{"_id":"$status",
                   "count":{"$count":{}},
                   }
         },
    ])
    badcount = 0
    goodcount = 0
    info = {}
    crawled_count = 0
    bad_crawl_count = 0
    for item in res:
        count = item["count"]
        st = item["_id"]
        print(st,count)
        if st == "good":
            goodcount += count
        if st != "frontlink" and st != "backlink": 
            crawled_count += count
            if st != "good":
                bad_crawl_count += count
        info[st] = count 
    info["crawled_count"] = crawled_count
    info["bad_crawl_count"] = bad_crawl_count
    baclink_cout = 0
    if "backlink" in info:
        backlink_count = info["backlink"]
    good_prob= 0
    if crawled_count > 0:
        good_prob = goodcount / crawled_count
    info["good_prob"] = good_prob
    print(">>>Domain Content")
    contentcol = db["content"]
    res = contentcol.aggregate([
        {"$match":{"host":hostname}},
        #{"$project": {"textsum":{"$sum":"$text_size"}}}
        {"$group":{"_id":None,
                   "text_size_sum":{"$sum":"$text_size"},
                   }
         },
    ])
    text_size = 0
    for item in res:
        text_size = item["text_size_sum"]
    good_document_characters = 0
    fetch_average_characters = 0
    if goodcount > 0:
        good_document_characters = text_size / goodcount
        fetch_average_characters = text_size / crawled_count
    info["total_good_characters"] = text_size
    info["average_good_characters"] = good_document_characters
    info["average_fetch_characters"] = fetch_average_characters
    domaincol = db["domain"]
    domaincol.update_one({"host":hostname},{"$set":info},upsert=True)
    res = domaincol.find_one({"host":hostname})
    print(res)

def sample_links(db,hostname,status,batch_size):
    print("Sampling links")
    linkcol = db["links"]
    res = linkcol.find({"host":hostname,"status": {"$not":{"$in":["frontlink","backlink"]}}})
    cl = LinkClassifier()
    crawled_links = list(res)
    crawled_count = len(crawled_links)
    prediction_accuracy = 0
    if crawled_count > CLASSIFIER_SET_SIZE:
        # train on crawled links
        trainset,testset = split_train(crawled_links)
        cl.train(trainset)
        prediction_accuracy = cl.test(testset)

    sample_set_size = SAMPLE_SET_SIZE
    res = linkcol.find({"host":hostname,"status": status})
    predicted_good = 0
    visitcounter = collections.Counter()
    good_links = []
    discover_links = []
    for item in res:
        link = item["url"]
        cll = cl.classify(link)
        if cll > 0: 
            good_links.append(link)
        features = get_link_features(link)
        discover_links.append(link)
        if features is None:
            continue
        for feature in features:
            visitcounter[feature] += 1
    mls = int(min(batch_size*(1- DISCOVER_LINK_RATIO),len(good_links)))
    random.shuffle(good_links)
    links = list(good_links[0:mls])
    numdiscover = len(discover_links)
    eval_discover_links = []
    for link in discover_links:
        features = get_link_features(link)
        prob = 0
        if features is not None:
            for feature in features:
                c = visitcounter[feature]
                prob -= math.log(c) / c
        eval_discover_links.append((link,prob))
    eval_discover_links.sort(key=lambda x: x[1],reverse=True)
    #print(eval_discover_links)
    mls = int(min(batch_size * DISCOVER_LINK_RATIO,len(eval_discover_links)))
    links += [l[0] for l in eval_discover_links[0:mls]]
    return list(set(links))

def domain_summary(db,hostname):
    linkcol = db["links"]
    #res = linkcol.distinct("hostname",{"hostname":hostname})
    
    # count links
    res = linkcol.aggregate([
        {"$group":{"_id":"$hostname","text_size_sum":{"$sum":"$text_size"}}},
    ])
    for item in res:
        print(item)


def createdb():
    myclient = pymongo.MongoClient(CONNECTION)
    db=myclient[DBNAME]
    linkcol = db["links"]
    linkcol.create_index("url",unique=True)
    linkcol.create_index("host")
    contentcol = db["content"]
    contentcol.create_index("url")
    contentcol.create_index("text_md5",unique=True)
    #contentcol.create_index({"paragraph_checksums":1})
    contentcol.create_index("host")
    htmlcol = db["html"]
    htmlcol.create_index("url")
    htmlcol.create_index("html_md5",unique=True)
    domaincol = db["domains"]
    domaincol.create_index("host",unique=True)
    domaincol.create_index("average_fetch_characters",unique=True)
    batchcol = db["batches"]
    batchcol.create_index("host")
    batchcol.create_index("created_at")

def parseurl(link):
    link,hostname = courlan.check_url(link)
    rawrules = trafilatura.fetch_url("https://"+ hostname + "/robots.txt")
    print(rawrules)
    rules = urllib.robotparser.RobotFileParser()
    rules.parse(rawrules.split("\n"))
    print(rules.can_fetch("*",link))
    print(rules.site_maps())
    print(rules.crawl_delay("*"))
    html = trafilatura.fetch_url(link,decode=True)
    get_bs_links(link,html)
    doc = trafilatura.bare_extraction(html)
    import pprint
    pprint.pprint(doc)
    internal_links, external_links = get_bs_links(link,html)
    print(internal_links)
    print(external_links)



def externaldomains(link):
    html = trafilatura.fetch_url(link,decode=True)
    external_links = courlan.extract_links(html,link,external_bool=True,language=LANGUAGE)
    domains = set()
    for l in external_links:
        r = courlan.check_url(l)
        if r is None:
            pass
        link,domain = r
        domains.add(domain)
    for d in domains:
        print(d)

def classify(start_link):
    myclient = pymongo.MongoClient(CONNECTION)
    db=myclient[DBNAME]
    start_link,hostname = courlan.check_url(start_link)
    cl = LinkClassifier()
    linkcol = db["links"]
    res = linkcol.find({"host":hostname,"status": {"$not":{"$in":["frontlink","backlink"]}}})
    trainset, testset = split_train(res)

    cl.train(trainset)
    cl.test(testset)

def visit(start_link):
    myclient = pymongo.MongoClient(CONNECTION)
    db=myclient[DBNAME]
    start_link,hostname = courlan.check_url(start_link)
    batch_size = BATCHSIZE
    rules = fetch_robot(hostname)
    # renew front links
    front_links = fetch_front_links(start_link,rules)
    index_links(db,front_links)
    # start crawling
    # frontlinks first
    links = sample_links(db,hostname,"frontlink",batch_size)
    if start_link not in links:
        links.insert(0,start_link)
    print("sampled")
    print(links)
    # index results
    print("Processing links")
    responses = []
    for link in links:
        responses.append(fetch_page(link))
    extracted_pages = extract_pages(links,responses)
    extracted_links = extract_links(links,responses,hostname,rules,"frontlink")
    index_links(db,extracted_links)
    index_pages(db,hostname,extracted_pages)
    link_summary(db,hostname)

